@@ -1,10 +1,9 @@
 from .utils import logger
-from .tree import Tree
+from .tree import Tree, ParseTree
 from .visitors import Transformer, Visitor, v_args, Discard, Transformer_NonRecursive
-from .visitors import InlineTransformer, inline_args   # XXX Deprecated
 from .exceptions import (ParseError, LexError, GrammarError, UnexpectedToken,
                          UnexpectedInput, UnexpectedCharacters, UnexpectedEOF, LarkError)
 from .lexer import Token
 from .lark import Lark
 
-__version__ = "0.11.3"
+__version__: str = "1.1.3"
